@@ -2,7 +2,6 @@ import logging
 import os
 import re
 import tempfile
-from collections import defaultdict
 from collections.abc import Mapping
 from concurrent.futures import (
     ProcessPoolExecutor,
@@ -15,12 +14,13 @@ from typing import Iterable, Optional
 
 from funcy import cached_property, first
 
-from dvc.exceptions import DownloadError, DvcException
+from dvc.exceptions import DownloadError, DvcException, UploadError
 from dvc.path_info import PathInfo
 from dvc.progress import Tqdm
 from dvc.repo.experiments.executor import ExperimentExecutor, LocalExecutor
 from dvc.scm.git import Git
 from dvc.stage.serialize import to_lockfile
+from dvc.tree.repo import RepoTree
 from dvc.utils import dict_sha256, env2bool, relpath
 from dvc.utils.fs import remove
 
@@ -53,6 +53,8 @@ class UnchangedExperimentError(DvcException):
 
 class BaselineMismatchError(DvcException):
     def __init__(self, rev, expected):
+        if hasattr(rev, "hexsha"):
+            rev = rev.hexsha
         rev_str = f"{rev[:7]}" if rev is not None else "dangling commit"
         super().__init__(
             f"Experiment derived from '{rev_str}', expected '{expected[:7]}'."
@@ -160,13 +162,25 @@ class Experiments:
     def _scm_checkout(self, rev):
         self.scm.repo.git.reset(hard=True)
         if self.scm.repo.head.is_detached:
-            # switch back to default branch
-            self.scm.repo.heads[0].checkout()
+            self._checkout_default_branch()
         if not Git.is_sha(rev) or not self.scm.has_rev(rev):
             self.scm.pull()
         logger.debug("Checking out experiment commit '%s'", rev)
         self.scm.checkout(rev)
 
+    def _checkout_default_branch(self):
+        # switch to default branch
+        git_repo = self.scm.repo
+        origin_refs = git_repo.remotes["origin"].refs
+        ref = origin_refs["HEAD"].reference
+        branch_name = ref.name.split("/")[-1]
+        if branch_name in git_repo.heads:
+            branch = git_repo.heads[branch_name]
+        else:
+            branch = git_repo.create_head(branch_name, ref)
+            branch.set_tracking_branch(ref)
+        branch.checkout()
+
     def _stash_exp(self, *args, params: Optional[dict] = None, **kwargs):
         """Stash changes from the current (parent) workspace as an experiment.
 
@@ -180,7 +194,9 @@ class Experiments:
         # patch user's workspace into experiments clone
         tmp = tempfile.NamedTemporaryFile(delete=False).name
         try:
-            self.repo.scm.repo.git.diff(patch=True, output=tmp)
+            self.repo.scm.repo.git.diff(
+                patch=True, full_index=True, binary=True, output=tmp
+            )
             if os.path.getsize(tmp):
                 logger.debug("Patching experiment workspace")
                 self.scm.repo.git.apply(tmp)
@@ -213,12 +229,7 @@ class Experiments:
 
     def _update_params(self, params: dict):
         """Update experiment params files with the specified values."""
-        from dvc.utils.serialize import (
-            dump_toml,
-            dump_yaml,
-            parse_toml_for_update,
-            parse_yaml_for_update,
-        )
+        from dvc.utils.serialize import MODIFIERS
 
         logger.debug("Using experiment params '%s'", params)
 
@@ -231,24 +242,18 @@ class Experiments:
                     dict_[key] = value
             return dict_
 
-        loaders = defaultdict(lambda: parse_yaml_for_update)
-        loaders.update({".toml": parse_toml_for_update})
-        dumpers = defaultdict(lambda: dump_yaml)
-        dumpers.update({".toml": dump_toml})
-
         for params_fname in params:
             path = PathInfo(self.exp_dvc.root_dir) / params_fname
-            with self.exp_dvc.tree.open(path, "r") as fobj:
-                text = fobj.read()
             suffix = path.suffix.lower()
-            data = loaders[suffix](text, path)
-            _update(data, params[params_fname])
-            dumpers[suffix](path, data)
+            modify_data = MODIFIERS[suffix]
+            with modify_data(path, tree=self.exp_dvc.tree) as data:
+                _update(data, params[params_fname])
 
     def _commit(self, exp_hash, check_exists=True, branch=True):
         """Commit stages as an experiment and return the commit SHA."""
         if not self.scm.is_dirty():
             raise UnchangedExperimentError(self.scm.get_rev())
+
         rev = self.scm.get_rev()
         exp_name = f"{rev[:7]}-{exp_hash}"
         if branch:
@@ -339,19 +344,26 @@ class Experiments:
                 for rev in revs
             }
 
-        # setup executors
+        logger.debug(
+            "Reproducing experiment revs '%s'",
+            ", ".join((rev[:7] for rev in to_run)),
+        )
+
+        # setup executors - unstash experiment, generate executor, upload
+        # contents of (unstashed) exp workspace to the executor tree
         executors = {}
         for rev, baseline_rev in to_run.items():
-            tree = self.scm.get_tree(rev)
-            repro_args, repro_kwargs = self._unpack_args(tree)
+            self._scm_checkout(baseline_rev)
+            self.scm.repo.git.stash("apply", rev)
+            repro_args, repro_kwargs = self._unpack_args()
             executor = LocalExecutor(
-                tree,
                 baseline_rev,
                 repro_args=repro_args,
                 repro_kwargs=repro_kwargs,
                 dvc_dir=self.dvc_dir,
                 cache_dir=self.repo.cache.local.cache_dir,
             )
+            self._collect_input(executor)
             executors[rev] = executor
 
         exec_results = self._reproduce(executors, **kwargs)
@@ -434,34 +446,56 @@ class Experiments:
 
         return result
 
+    def _collect_input(self, executor: ExperimentExecutor):
+        """Copy (upload) input from the experiments workspace to the executor
+        tree.
+        """
+        logger.debug("Collecting input for '%s'", executor.tmp_dir)
+        repo_tree = RepoTree(self.exp_dvc)
+        self._process(
+            executor.tree,
+            self.exp_dvc.tree,
+            executor.collect_files(self.exp_dvc.tree, repo_tree),
+        )
+
     def _collect_output(self, executor: ExperimentExecutor):
         """Copy (download) output from the executor tree into experiments
         workspace.
         """
-        from dvc.cache.local import _log_exceptions
-
         logger.debug("Collecting output from '%s'", executor.tmp_dir)
-        dest_tree = self.exp_dvc.tree
-        src_tree = executor.tree
+        self._process(
+            self.exp_dvc.tree,
+            executor.tree,
+            executor.collect_output(),
+            download=True,
+        )
+
+    @staticmethod
+    def _process(dest_tree, src_tree, collected_files, download=False):
+        from dvc.cache.local import _log_exceptions
 
         from_infos = []
         to_infos = []
         names = []
-        for from_info in executor.collect_output():
+        for from_info in collected_files:
             from_infos.append(from_info)
             fname = from_info.relative_to(src_tree.path_info)
             names.append(str(fname))
-            to_info = dest_tree.path_info / fname
             to_infos.append(dest_tree.path_info / fname)
-            logger.debug(f"from '{from_info}' to '{to_info}'")
         total = len(from_infos)
 
-        func = partial(
-            _log_exceptions(dest_tree.download, "download"),
-            dir_mode=dest_tree.dir_mode,
-            file_mode=dest_tree.file_mode,
-        )
-        with Tqdm(total=total, unit="file", desc="Downloading") as pbar:
+        if download:
+            func = partial(
+                _log_exceptions(src_tree.download, "download"),
+                dir_mode=dest_tree.dir_mode,
+                file_mode=dest_tree.file_mode,
+            )
+            desc = "Downloading"
+        else:
+            func = partial(_log_exceptions(dest_tree.upload, "upload"))
+            desc = "Uploading"
+
+        with Tqdm(total=total, unit="file", desc=desc) as pbar:
             func = pbar.wrap_fn(func)
             # TODO: parallelize this, currently --jobs for repro applies to
             # number of repro executors not download threads
@@ -469,7 +503,9 @@ class Experiments:
                 fails = sum(dl_executor.map(func, from_infos, to_infos, names))
 
         if fails:
-            raise DownloadError(fails)
+            if download:
+                raise DownloadError(fails)
+            raise UploadError(fails)
 
     @scm_locked
     def checkout_exp(self, rev):
@@ -482,12 +518,14 @@ class Experiments:
         self._scm_checkout(rev)
 
         tmp = tempfile.NamedTemporaryFile(delete=False).name
-        self.scm.repo.head.commit.diff("HEAD~1", patch=True, output=tmp)
+        self.scm.repo.head.commit.diff(
+            "HEAD~1", patch=True, full_index=True, binary=True, output=tmp
+        )
 
         dirty = self.repo.scm.is_dirty()
         if dirty:
             logger.debug("Stashing workspace changes.")
-            self.repo.scm.repo.git.stash("push")
+            self.repo.scm.repo.git.stash("push", "--include-untracked")
 
         try:
             if os.path.getsize(tmp):
@@ -519,10 +557,23 @@ class Experiments:
         # to prefer the applied experiment changes over stashed workspace
         # changes. git stash doesn't support merge strategy parameters, but we
         # can do it ourselves with checkout/reset.
+        from git.exc import GitCommandError
+
         logger.debug("Unstashing workspace changes.")
-        self.repo.scm.repo.git.checkout("--ours", "stash@{0}", "--", ".")
-        self.repo.scm.repo.git.reset("HEAD")
-        self.repo.scm.repo.git.stash("drop", "stash@{0}")
+        git_repo = self.repo.scm.repo.git
+
+        # stage workspace changes, then apply stashed changes on top
+        git_repo.add(A=True)
+        try:
+            git_repo.stash("apply", "stash@{0}")
+        except GitCommandError:
+            # stash apply will return error code on merge conflicts,
+            # prefer workspace changes over stash changes
+            git_repo.checkout("--ours", "--", ".")
+
+        # unstage changes and drop the stash entry
+        git_repo.reset("HEAD")
+        git_repo.stash("drop", "stash@{0}")
 
     def checkout(self, *args, **kwargs):
         from dvc.repo.experiments.checkout import checkout
